package org.bitbucket.lp.maven.plugins.svuid;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.ResolutionScope;

/**
 * This mojo lists the classes that declare serialVersionUID fields. It also
 * prints the lines where the declarations occur and the value associated to the
 * fields.
 * 
 * @author lpellegr
 */
@Mojo(name = "list", defaultPhase = LifecyclePhase.NONE, requiresOnline = false, requiresDependencyResolution = ResolutionScope.NONE, requiresDirectInvocation = true, threadSafe = false)
public class ListMojo extends SvuidMojo {

    private static final String COLUMN_TITLE_CLASSNAME = "Class";

    private static final String COLUMN_TITLE_LINE = "Line";

    private static final String COLUMN_TITLE_VALUE = "Value";

    private static final int COLUMNS_SPACE_SIZE = 2;

    /**
     * {@inheritDoc}
     */
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        Map<File, List<SerialVersionUidEntry>> entries =
                findFilesDeclaringSerialVersionUID(super.srcFolder);

        Log log = super.getLog();

        if (entries.isEmpty()) {
            log.info(NO_FILE_FOUND_MSG);
            return;
        }

        SvuidMaxLength lengths =
                findAttributesMaxLength(entries, super.executionDir);
        lengths.maxLineLength =
                Math.max(lengths.maxLineLength, COLUMN_TITLE_LINE.length());
        lengths.maxValueLength =
                Math.max(lengths.maxValueLength, COLUMN_TITLE_VALUE.length());

        StringBuilder buffer = new StringBuilder();
        buffer.append(COLUMN_TITLE_CLASSNAME);
        append(buffer, ' ', lengths.maxFilenameLength
                - COLUMN_TITLE_CLASSNAME.length() + COLUMNS_SPACE_SIZE);
        buffer.append(COLUMN_TITLE_LINE);
        append(buffer, ' ', positiveOrZero(lengths.maxLineLength
                - COLUMN_TITLE_LINE.length())
                + COLUMNS_SPACE_SIZE);
        buffer.append(COLUMN_TITLE_VALUE);
        append(buffer, ' ', positiveOrZero(lengths.maxValueLength
                - COLUMN_TITLE_VALUE.length()));

        log.info(buffer.toString());
        buffer = new StringBuilder();

        for (Entry<File, List<SerialVersionUidEntry>> entry : entries.entrySet()) {
            for (int i = 0; i < entry.getValue().size(); i++) {
                if (i == 0) {
                    String filename =
                            shorten(entry.getKey(), super.executionDir);
                    buffer.append(filename);
                    append(buffer, ' ', lengths.maxFilenameLength
                            - filename.length() + COLUMNS_SPACE_SIZE);
                } else {
                    append(buffer, ' ', lengths.maxFilenameLength
                            + COLUMNS_SPACE_SIZE);
                }

                String entryLine =
                        Integer.toString(entry.getValue().get(i).line);
                String entryValue =
                        Long.toString(entry.getValue().get(i).value);

                append(buffer, ' ', lengths.maxLineLength - entryLine.length());
                buffer.append(entryLine);
                append(buffer, ' ', COLUMNS_SPACE_SIZE);
                append(buffer, ' ', lengths.maxValueLength
                        - entryValue.length());
                buffer.append(entryValue);
                append(buffer, ' ', COLUMNS_SPACE_SIZE);

                log.info(buffer.toString());
                buffer = new StringBuilder();
            }
        }

        log.info(buffer.toString());

    }

    private static final int positiveOrZero(int v) {
        if (v < 0) {
            return 0;
        }

        return v;
    }

    private static final void append(StringBuilder buffer, char c, int n) {
        for (int i = 0; i < n; i++) {
            buffer.append(c);
        }
    }

    private static final SvuidMaxLength findAttributesMaxLength(Map<File, List<SerialVersionUidEntry>> entries,
                                                                File executionDir) {
        SvuidMaxLength result = new SvuidMaxLength();

        for (Entry<File, List<SerialVersionUidEntry>> entry : entries.entrySet()) {
            int filenameLength = shorten(entry.getKey(), executionDir).length();

            if (filenameLength > result.maxFilenameLength) {
                result.maxFilenameLength = filenameLength;
            }

            for (SerialVersionUidEntry svuidEntry : entry.getValue()) {
                int lineLength = Integer.toString(svuidEntry.line).length();
                int valueLength = Long.toString(svuidEntry.value).length();

                if (lineLength > result.maxLineLength) {
                    result.maxLineLength = lineLength;
                }

                if (valueLength > result.maxValueLength) {
                    result.maxValueLength = valueLength;
                }
            }
        }

        return result;
    }

    private static final String shorten(File filePath, File executionDir) {
        return filePath.toString().replace(
                executionDir.toString() + File.separatorChar, "");
    }

}
