/**
 * Copyright (c) 2011-2012 INRIA.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>
 **/
package org.bitbucket.lp.maven.plugins.svuid;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.Parameter;
import org.bitbucket.lp.maven.plugins.svuid.transformers.Transformer;

/**
 * Abstraction used by any SvuidMojo to factorize code and share common fields.
 * 
 * @author lpellegr
 */
public abstract class SvuidMojo extends AbstractMojo {

    protected static final Pattern SVUID_PATTERN =
            Pattern.compile("((public|private|protected)?\\p{Blank}+static\\p{Blank}+final\\p{Blank}+long\\p{Blank}+serialVersionUID\\p{Blank}+=\\p{Blank}+)([-]?\\d+L)(\\p{Blank}*;)");

    protected static final String NO_FILE_FOUND_MSG =
            "No class declaring a serialVersionUID field";

    @Parameter(defaultValue = "${project.basedir}", required = true)
    protected File srcFolder;

    @Parameter(defaultValue = "${session.executionRootDirectory}", required = true)
    protected File executionDir;

    protected static final Map<File, List<SerialVersionUidEntry>> findFilesDeclaringSerialVersionUID(File dirPath)
            throws MojoExecutionException {
        Collection<File> files =
                FileUtils.listFiles(dirPath, new String[] {"java"}, true);

        Map<File, List<SerialVersionUidEntry>> filesMatching =
                new HashMap<File, List<SerialVersionUidEntry>>();

        for (File file : files) {
            List<SerialVersionUidEntry> entries = findSerialVersionUID(file);

            if (!entries.isEmpty()) {
                filesMatching.put(file, entries);
            }
        }

        return filesMatching;
    }

    private static final List<SerialVersionUidEntry> findSerialVersionUID(File file)
            throws MojoExecutionException {
        Matcher matcher = SVUID_PATTERN.matcher("");

        LineNumberReader lineReader = null;
        List<SerialVersionUidEntry> result =
                new ArrayList<SerialVersionUidEntry>();
        try {
            lineReader = new LineNumberReader(new FileReader(file));
            String line = null;

            while ((line = lineReader.readLine()) != null) {
                matcher.reset(line);

                if (matcher.find()) {
                    String svuidValue = matcher.group(3);
                    if (svuidValue.endsWith("L")) {
                        svuidValue =
                                svuidValue.substring(0, svuidValue.length() - 1);
                    }

                    result.add(new SerialVersionUidEntry(
                            lineReader.getLineNumber(),
                            Long.parseLong(svuidValue)));
                }
            }
        } catch (FileNotFoundException e) {
            throw new MojoExecutionException(e.getMessage(), e);
        } catch (IOException e) {
            throw new MojoExecutionException(e.getMessage(), e);
        } finally {
            try {
                if (lineReader != null) {
                    lineReader.close();
                }
            } catch (IOException ex) {
                throw new MojoExecutionException(ex.getMessage(), ex);
            }
        }

        return result;
    }

    protected static final void replaceSerialVersionUidFields(File file,
                                                              Transformer transformer)
            throws MojoExecutionException {
        BufferedReader reader = null;
        BufferedWriter writer = null;

        Matcher matcher = SVUID_PATTERN.matcher("");

        File tmpFile = null;
        try {
            tmpFile = new File(file.toString() + ".svuid");

            reader = new BufferedReader(new FileReader(file));
            writer = new BufferedWriter(new FileWriter(tmpFile));

            String line = null;

            while ((line = reader.readLine()) != null) {
                matcher.reset(line);

                String replacementString = line;

                if (matcher.find()) {
                    String stringValue = matcher.group(3);
                    if (stringValue.endsWith("L")) {
                        stringValue =
                                stringValue.substring(
                                        0, stringValue.length() - 1);
                    }

                    long newValue =
                            transformer.apply(Long.parseLong(stringValue));

                    if (newValue != -1) {
                        replacementString =
                                line.replaceAll(SVUID_PATTERN.pattern(), "$1"
                                        + newValue + "L$4");
                    }
                }

                writer.write(replacementString);
                writer.write('\n');
            }
        } catch (FileNotFoundException e) {
            throw new MojoExecutionException(e.getMessage(), e);
        } catch (IOException e) {
            throw new MojoExecutionException(e.getMessage(), e);
        } finally {
            try {
                if (reader != null) {
                    reader.close();
                }
                if (writer != null) {
                    writer.close();
                }

                if (tmpFile != null) {
                    tmpFile.renameTo(file);
                }
            } catch (IOException ex) {
                throw new MojoExecutionException(ex.getMessage(), ex);
            }
        }
    }

    protected static final class SvuidMaxLength {

        public int maxFilenameLength;

        public int maxLineLength;

        public int maxValueLength;

    }

    protected static final class SerialVersionUidEntry {

        public final int line;

        public final long value;

        public SerialVersionUidEntry(int line, long value) {
            this.line = line;
            this.value = value;
        }

    }

}
